/*
 * Lww.c - Lww Widget
 *
 * $Id: Lww.c,v 1.2 1997/12/29 14:52:36 k-chinen Exp k-chinen $
 *
 */

#include <stdio.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/XawInit.h>

#include "LwwP.h"

static XtResource resources[] = {
   {XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
      XtOffsetOf(LwwRec, core.width), XtRImmediate,
      (XtPointer) 100},
   {XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
      XtOffsetOf(LwwRec, core.height), XtRImmediate,
      (XtPointer) 100},
   {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
      XtOffsetOf(LwwRec, core.background_pixel), XtRString,
      XtDefaultForeground},
   {XtNborderWidth, XtCBorderWidth, XtRDimension, sizeof(Dimension),
      XtOffsetOf(LwwRec, core.border_width), XtRImmediate, (XtPointer)0},
   {XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer),
      XtOffsetOf(LwwRec, lww.lww_action), XtRCallback, NULL},
   {XtNvisual, XtCVisual, XtRVisual, sizeof(Visual*),
      XtOffsetOf(LwwRec, lww.visual), XtRImmediate, CopyFromParent}
};

static void LwwAction( /* Widget, XEvent*, String*, Cardinal */ );

static XtActionsRec actionsList[] =
{
  {"LwwAction",      LwwAction},
};

#define SuperClass (&simpleClassRec)

static void Realize();
static void DummyExpose();

static int debug=0;


LwwClassRec lwwClassRec = {
   {
/* core class fields */
    /* superclass         */   (WidgetClass) SuperClass,
    /* class name         */   "Lww",
    /* size               */   sizeof(LwwRec),
    /* class initialize   */   XawInitializeWidgetSet,
    /* class_part_init    */   NULL,
    /* class_inited       */   FALSE,
    /* initialize         */   NULL,
    /* initialize_hook    */   NULL,
#if 0
    /* realize            */   XtInheritRealize,
#else
    /* realize            */   Realize,
#endif
    /* actions            */   actionsList,
    /* num_actions        */   XtNumber(actionsList),
    /* resources          */   resources,
    /* resource_count     */   XtNumber(resources),
    /* xrm_class          */   NULLQUARK,
    /* compress_motion    */   TRUE,
    /* compress_exposure  */   TRUE,
    /* compress_enterleave*/   TRUE,
    /* visible_interest   */   FALSE,
    /* destroy            */   NULL,
    /* resize             */   NULL,
#if 1
    /* expose             */   NULL,
#else
    /* expose             */   DummyExpose,
#endif
    /* set_values         */   NULL,
    /* set_values_hook    */   NULL,
    /* set_values_almost  */   XtInheritSetValuesAlmost,
    /* get_values_hook    */   NULL,
    /* accept_focus       */   NULL,
    /* version            */   XtVersion,
    /* callback_private   */   NULL,
    /* tm_table           */   NULL,
    /* query_geometry     */   XtInheritQueryGeometry,
    /* display_accelerator*/   XtInheritDisplayAccelerator,
    /* extension          */   NULL
   },
/* Simple class fields initialization */
   {
    /* change_sensitive   */   XtInheritChangeSensitive
   },
/* Lww class fields initialization */
   {
    /* not used           */   0
   }
};

WidgetClass lwwWidgetClass = (WidgetClass) &lwwClassRec;

static void LwwAction( widget, event, params, num_params )
    Widget widget;
    XEvent *event;
    String *params;
    Cardinal *num_params;
{
    LwwCallDataRec call_data;

    call_data.event = event;
    call_data.params = params;
    call_data.num_params = *num_params;

    XtCallCallbacks( widget, XtNcallback, (XtPointer)&call_data );
}


static
void
Realize (gw, valueMask, attrs)
    LwwWidget gw;
    XtValueMask *valueMask;
    XSetWindowAttributes *attrs;
{
    Colormap cmap;
    Visual *vis;
    int     depth;

    vis = gw->lww.visual;
    if(vis==NULL) {
        fprintf(stderr, "vis is NULL\n", vis, vis);
        XtInheritRealize((Widget)gw, valueMask, attrs);

        return;
    }

    depth = gw->core.depth;

    if(debug>=10)
        fprintf(stderr, "vis %p %d (depth %d)\n", vis, vis, depth);
    if(vis->class == PseudoColor) {
        if(debug>=10)
            fprintf(stderr, "\tPseudoColor\n");
        cmap = XCreateColormap(XtDisplay(gw),
                gw->core.parent->core.window,
                vis, AllocAll);
    }
    if(vis->class == TrueColor) {
        if(debug>=10)
            fprintf(stderr, "\tTrueColor\n");
        cmap = XCreateColormap(XtDisplay(gw),
                gw->core.parent->core.window,
                vis, AllocNone);
    }
    if(vis->class == DirectColor) {
        fprintf(stderr, "\tDirectColor\n");
        cmap = XCreateColormap(XtDisplay(gw),
                gw->core.parent->core.window,
                vis, AllocNone);
    }

    *valueMask |= CWColormap;
    *valueMask |= CWBackPixel;
    attrs->colormap = cmap;
    attrs->background_pixel = gw->core.background_pixel;

    gw->core.window = XCreateWindow(
            XtDisplay(gw),
            gw->core.parent->core.window,
            gw->core.x, gw->core.y, gw->core.width, gw->core.height,
            gw->core.border_width,
            depth, InputOutput, vis,
            *valueMask, attrs);

    XFlush(XtDisplay(gw));
    if(debug>=10) {
        fprintf(stderr, "done\n");
        fflush(stderr);
    }   
}

static
void
DummyExpose(gw, ev, reg)
    Widget gw;
    XEvent* ev;
    Region reg;
{
    GC          gc;
    int         x, y;
    int         cv;

#if 0
    unsigned long gc_value_mask;
    XGCValues   *gc_values;

    gc_value_mask = 0;
    gc_values = NULL;
    gc = XCreateGC(XtDisplay(gw), XtWindow(gw),
        gc_value_mask, gc_values);
#else
    XtGCMask    gc_value_mask;
    XGCValues   gc_values;

    gc_value_mask = GCFunction;
    gc_values.function = GXcopy;
    gc = XtGetGC(gw, gc_value_mask, &gc_values);
#endif

    XSetFunction(XtDisplay(gw), gc, GXcopy);

#if 0
    for(y=0;y<256;y++) {
        for(x=0;x<256;x++) {
            cv = (x<<8)+(y);
            fprintf(stderr, "%3d %2x / %3d %2x - %8x\n",
                x, x, y, y, cv);
            XSetForeground(XtDisplay(gw), gc, cv);
            XDrawPoint(XtDisplay(gw), XtWindow(gw), gc, x, y);
        }
    }
#endif

#if 1
    for(y=0;y<256;y+=16) {
        for(x=0;x<256;x+=16) {
            cv = (x<<8)+(y);
            fprintf(stderr, "%3d %2x / %3d %2x - %06x\n",
                x, x, y, y, cv);
            XSetForeground(XtDisplay(gw), gc, cv);
            XFillRectangle(XtDisplay(gw), XtWindow(gw), gc, x, y, 16, 16);
        }
    }
#endif

    XFlush(XtDisplay(gw));

#if 1
    XtReleaseGC(gw, gc);
#endif
}


