/*
 * get_data.c - get data (image header and body)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include "hook.h"
#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"

#ifdef  HAVE_TYPEDEF_FD_SET
typedef fd_set          fdset_t;
#else
typedef struct fd_set   fdset_t;
#endif  /* HAVE_FDSET_T */



/*
 * isreadable - check specified FD is readable or not.
 *      if sec<0        wait with blocking
 *
 *
 * return:
 *      1   readable
 *      0   not readable
 *     -1   error
 *
 */

int
isreadable_via_select(int fd, int sec, int usec)
{
    fdset_t fdset, rfdset;
    struct timeval timeout;
    struct timeval *ptimeout;
    int chk;

    FD_ZERO(&fdset);
    FD_SET(fd, &fdset);

    if(sec<0) {
        ptimeout = NULL;
    }
    else {
        ptimeout = &timeout;

        timeout.tv_sec  =  sec;
        if(usec<0) {
            timeout.tv_usec = 0;
        }
        else {
            timeout.tv_usec = usec;
        }
    }

retry:
    rfdset = fdset;
    chk = select(fd+1, &rfdset, NULL, NULL, ptimeout);

#if 0
    dump_fdset(fd+1, &fdset, &rfdset, NULL, NULL);
#endif
    if(chk<0) {
        if(errno==EINTR) {
            goto retry;
        }
        else {
            Error(" #isreadable:fd# %d, errno %d# ", fd, errno);
            return -1;
        }
    }

    if(FD_ISSET(fd, &rfdset)) {
#if 0
        Trace("readable ");
#endif
        return 1;
    }
    else {
#if 0
        Trace("unreadable ");
#endif
        return 0;
    }
}



#if 0
int
get_data2(int datafd, char **buff,
    sentry_shmrec *info, sentry_shmrec *dest)
{
    struct sockaddr_in from;
    int fromlen;
    char packet[8192];
    int nread;
    sentry_protocol3_image_hdr *hdr;
    int chk;
    struct timeval image_tstamp;
    int image_format, image_length, image_width, image_height;
    struct timeval curtime;
    time_t now;
    int recvlen = -1;

    /*
     * receive packets 
    */
    while (1) {
        fromlen = sizeof(from);
        /* bzero(packet, sizeof(packet));*/
        gettimeofday(&curtime, NULL);

        nread = recvfrom(datafd, packet, sizeof(packet),
                    0, (struct sockaddr *)&from, &fromlen);
    
        if (nread < 0) {
          if (errno == EINTR) 
            continue;

          perror("get_data: fail recvfrom()");
          return 1;
        }

        if (debug>=100)
            fprintf(stderr, "get_data: receive %d bytes from %s:%d\n",
                nread, inet_ntoa(from.sin_addr), htons(from.sin_port));
    
        /*
         * header check
         */
        if ((nread -= SENTRY_IMAGE_HEADER_LEN) < 0) {
            fprintf(stderr, "get_data: header length is wrong.\n");
            return 1;
        }

        hdr = (sentry_protocol3_image_hdr *)packet;
    
        if (hdr->hd_ver != SENTRY_PROTOCOL_VERSION) {
          fprintf(stderr, "get_data: header version is wrong.\n");
          return 1;
        }

#if 0
    if (debug) {
      static int ac=0;
      char filename[256];

      sprintf(filename, "x-sample%d.jpg", ac);
      ac = (ac+1) % 7;
      dump_to_file(filename, &packet[SENTRY_IMAGE_HEADER_LEN], nread);
    }
#endif

        chk = udp_reassemble(hdr,
                &packet[SENTRY_IMAGE_HEADER_LEN], nread, buff);

#if 0
    if (debug && chk>0) {
      static int bc=0;
      char filename[256];

      sprintf(filename, "c-sample%d.jpg", bc);
      bc = (bc+1) % 7;
      dump_to_file(filename, *buff, chk);
    }
#endif

        if (chk > 0) {
            if(debug>=1000) {
                fprintf(stdout, ".\n");
            }

            recvlen = chk;
            break;
        }

        if (chk == 0) {
            /* need more udp packets, to construct image frame */
            if(debug>=1000) {
                fprintf(stdout, ".");
            }
            continue;
        } 
        else /* chk < 0 */
        {
            return 1;
        }

    } /* end-of-while */

    if(recvlen<0) {
        fprintf(stderr, "negative length data recieved\n");
        goto out;
    }
    if(dest->entry.header.length>0 && recvlen>dest->entry.header.body_len) {
        fprintf(stderr, "length in data %d is larger than storage %d\n",
            dest->entry.header.length, dest->entry.header.body_len);
        goto out;
    }
    if(recvlen>0 && recvlen>dest->entry.header.body_len) {
        fprintf(stderr,
            "length of received data %d is larger than storage %d\n",
            recvlen, dest->entry.header.body_len);
        goto out;
    }

    dest->entry.header.send_sec  = ntohl(hdr->hd_tstamp.tv_sec);
    dest->entry.header.send_usec = ntohl(hdr->hd_tstamp.tv_usec);
    dest->entry.header.recv_sec  = curtime.tv_sec;
    dest->entry.header.recv_usec = curtime.tv_usec;
    dest->entry.header.format   = ntohl(hdr->format);
    dest->entry.header.corder   = ntohl(hdr->color);
    dest->entry.header.length   = ntohl(hdr->size);
    dest->entry.header.width    = ntohl(hdr->width);
    dest->entry.header.height   = ntohl(hdr->height);
    if(debug>=100) {
        printf("  pos %p, length %d/%d, corder '%s'(%d)\n",
            dest->entry.header.body,
            dest->entry.header.length, dest->entry.header.body_len,
            assoc_v(color_alist, dest->entry.header.corder),
            dest->entry.header.corder
            );
    }
    memcpy(dest->entry.header.body, *buff, dest->entry.header.length);
    info->info.last_update = now;

 if(debug>=300) {
    fprintf(stderr,
        "packet %d.%06d; stored issued %d.%06d received %d.%06d\n",
        ntohl(hdr->hd_tstamp.tv_sec), ntohl(hdr->hd_tstamp.tv_usec),
        dest->entry.header.send_sec, dest->entry.header.send_usec,
        dest->entry.header.recv_sec, dest->entry.header.recv_usec);
 }
  
#if 0
    dump_rgb_to_file("/var/tmp/x.ppm", dest->entry.header.body,
        dest->entry.header.width, dest->entry.header.height);
#endif
#if 0
    dump_to_file("/var/tmp/x.jpg", dest->entry.header.body,
        dest->entry.header.length);
#endif

out:

    return 0;
}

#endif



int
get_data3(int datafd, char **buff,
    sentry_shmrec *info, sentry_imgrec *dest)
{
    struct sockaddr_in from;
    int fromlen;
    char packet[8192];
    int nread;
    sentry_protocol3_image_hdr *hdr;
    int chk;
    struct timeval image_tstamp;
    int image_format, image_length, image_width, image_height;
    struct timeval curtime;
    time_t now;
    int recvlen = -1;

    /*
     * receive packets 
    */
    while (1) {
        fromlen = sizeof(from);
        /* bzero(packet, sizeof(packet));*/
        gettimeofday(&curtime, NULL);

        nread = recvfrom(datafd, packet, sizeof(packet),
                    0, (struct sockaddr *)&from, &fromlen);
    
        if (nread < 0) {
          if (errno == EINTR) 
            continue;

          perror("get_data: fail recvfrom()");
          return 1;
        }

        if (debug>=100)
            fprintf(stderr, "get_data: receive %d bytes from %s:%d\n",
                nread, inet_ntoa(from.sin_addr), htons(from.sin_port));
    
        /*
         * header check
         */
        if ((nread -= SENTRY_IMAGE_HEADER_LEN) < 0) {
            fprintf(stderr, "get_data: header length is wrong.\n");
            return 1;
        }

        hdr = (sentry_protocol3_image_hdr *)packet;
    
        if (hdr->hd_ver != SENTRY_PROTOCOL_VERSION) {
          fprintf(stderr, "get_data: header version is wrong.\n");
          return 1;
        }

#if 0
    if (debug) {
      static int ac=0;
      char filename[256];

      sprintf(filename, "x-sample%d.jpg", ac);
      ac = (ac+1) % 7;
      dump_to_file(filename, &packet[SENTRY_IMAGE_HEADER_LEN], nread);
    }
#endif

        chk = udp_reassemble(hdr,
                &packet[SENTRY_IMAGE_HEADER_LEN], nread, buff);

#if 0
    if (debug && chk>0) {
      static int bc=0;
      char filename[256];

      sprintf(filename, "c-sample%d.jpg", bc);
      bc = (bc+1) % 7;
      dump_to_file(filename, *buff, chk);
    }
#endif

        if (chk > 0) {
            if(debug>=1000) {
                fprintf(stdout, ".\n");
            }

            recvlen = chk;
            break;
        }

        if (chk == 0) {
            /* need more udp packets, to construct image frame */
            if(debug>=1000) {
                fprintf(stdout, ".");
            }
            continue;
        } 
        else /* chk < 0 */
        {
            return 1;
        }

    } /* end-of-while */

    if(recvlen<0) {
        fprintf(stderr, "negative length data recieved\n");
        goto out;
    }
    if(dest->length>0 && recvlen>dest->body_len) {
        fprintf(stderr, "length in data %d is larger than storage %d\n",
            dest->length, dest->body_len);
        goto out;
    }
    if(recvlen>0 && recvlen>dest->body_len) {
        fprintf(stderr,
            "length of received data %d is larger than storage %d\n",
            recvlen, dest->body_len);
        goto out;
    }

    dest->send_sec   = ntohl(hdr->hd_tstamp.tv_sec);
    dest->send_usec = ntohl(hdr->hd_tstamp.tv_usec);
    dest->recv_sec   = curtime.tv_sec;
    dest->recv_usec = curtime.tv_usec;
    dest->format    = ntohl(hdr->format);
    dest->corder    = ntohl(hdr->color);
    dest->length    = ntohl(hdr->size);
    dest->width     = ntohl(hdr->width);
    dest->height    = ntohl(hdr->height);
    if(debug>=100) {
        printf("  pos %p, length %d/%d, format '%s'(%d) corder '%s'(%d)\n",
            dest->body,
            dest->length, dest->body_len,
            assoc_v(format_alist, dest->format),
            dest->format,
            assoc_v(color_alist, dest->corder),
            dest->corder
            );
    }
    memcpy(dest->body, *buff, dest->length);
    info->info.last_update = now;

 if(debug>=300) {
    fprintf(stderr,
        "packet %d.%06d; stored issued %d.%06d received %d.%06d\n",
        ntohl(hdr->hd_tstamp.tv_sec), ntohl(hdr->hd_tstamp.tv_usec),
        dest->send_sec, dest->send_usec,
        dest->recv_sec, dest->recv_usec);
 }
  
#if 0
    dump_rgb_to_file("/var/tmp/x.ppm", dest->body,
        dest->width, dest->height);
#endif
#if 0
    dump_to_file("/var/tmp/x.jpg", dest->body,
        dest->length);
#endif

out:

    return 0;
}




