/*
 * 
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <sys/types.h>

#include <time.h>
#include <sys/time.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <syslog.h>
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"


extern int debug;
extern int verbose;

void init_shm();
void show_sentry_shm();
void save_thru(struct timeval, int, int, char**);

key_t   sentry_shmkey = SENTRY_SHMKEY;
int     sentry_shm_chunk_radix = SENTRY_SHM_CHUNK_RADIX;
int     sentry_shmid;
sentry_shmrec   *sentry_shm_chunk;
sentry_shmrec   *sentry_shm_info;
/*
extern key_t   sentry_shmkey;
extern int     sentry_shm_chunk_radix;
extern int     sentry_shmid;
extern sentry_shmrec   *sentry_shm_chunk;
extern sentry_shmrec   *sentry_shm_info;
*/

int pool_inited=0;

int clear_shm(void);

int
probe_shm(void)
{
    sentry_shmrec x;
    char *h, *b;

    h = (char*)&x;
    b = (char*)&x.entry.body_area;

    printf("x %p, x.body_area %p, diff %d\n", h, b, b-h);

    if(b-h != SENTRY_SHM_HEADER_SIZE) {
        return 1;
    }
    return 0;
}

void init_shm(int xkey)
{
  int chk;

  fprintf(stderr, "init_shm:\n");

  if(pool_inited) {
     return;
  }

  chk = probe_shm();
  if(chk) {
    fprintf(stderr,
        "ERROR: fatal error was found in offset of sentry_shmrec\n");
    exit(13);
  }

  pool_inited = 1;

  if(xkey==-1) {
#if 0
  sentry_shmkey = SENTRY_SHMKEY;
#endif
  }
  else {
    fprintf(stderr, "special key %d\n", xkey);
    sentry_shmkey = xkey;
  }
  syslog(LOG_NOTICE, "key %d", sentry_shmkey);

  sentry_shm_chunk_radix = SENTRY_SHM_CHUNK_RADIX;


  if ((sentry_shmid = 
       make_new_pool(sentry_shmkey, SENTRY_SHM_BLOCK_SIZE*(sentry_shm_chunk_radix+1))) < 0) {
    perror("init_shm: cannot make shm");
    exit (1);
  }
  if (debug) fprintf(stderr, "ok0: sentry_shmid=%d\n", sentry_shmid);

  if ((sentry_shm_chunk = (sentry_shmrec *)shmat(sentry_shmid, NULL, 0)) == (sentry_shmrec *)-1) {
    perror("init_shm: fail to shmat()");
    exit (1);
  }
  if (debug) fprintf(stderr, "ok1: sentry_shm_chunk=%d\n", sentry_shm_chunk);

  sentry_shm_info = &sentry_shm_chunk[0];
  bzero((char *)sentry_shm_info->info.magick, 16);
  strcpy(sentry_shm_info->info.magick, SENTRY_SHM_MAGIC);
  bzero((char *)sentry_shm_info->info.ver, 0);
/*
  strcpy(sentry_shm_info->info.ver, SENTRY_SHM_VERSION);
*/
  sprintf(sentry_shm_info->info.ver, "%d", SENTRY_SHM_VERSION);

  sentry_shm_info->info.last_update = -1;
  sentry_shm_info->info.radix = sentry_shm_chunk_radix;
  
  show_pool_stat(sentry_shmid);
  show_sentry_shm();

  clear_shm();
}


int
clear_shm(void)
{
  int i;
  int pos;
  int body_area_size;

  body_area_size = sizeof(sentry_shm_chunk[0].entry.body_area);
  if(debug>=100) {
    fprintf(stderr, "body_area_size = %d\n", body_area_size);
  }

  for(i=0;i<sentry_shm_chunk_radix;i++) {
    pos = i+1;
    sentry_shm_chunk[pos].entry.header.recv_sec = -1;
    sentry_shm_chunk[pos].entry.header.recv_usec = -1;
    sentry_shm_chunk[pos].entry.header.send_sec = -1;
    sentry_shm_chunk[pos].entry.header.send_usec = -1;
/*
    sentry_shm_chunk[pos].entry.header.length = csize;
*/
    sentry_shm_chunk[pos].entry.header.length = 0;
    sentry_shm_chunk[pos].entry.header.format = SEN_F_NONE;
    sentry_shm_chunk[pos].entry.header.width  = -1;
    sentry_shm_chunk[pos].entry.header.height = -1;
    sentry_shm_chunk[pos].entry.header.corder = SEN_C_NONE;
    sentry_shm_chunk[pos].entry.header.body_len = body_area_size;
    sentry_shm_chunk[pos].entry.header.body =
        &sentry_shm_chunk[pos].entry.body_area;
/*
    memset(&sentry_shm_chunk[pos].entry.body, 0xcc, csize);
*/
/*
    memset(&sentry_shm_chunk[pos].entry.body_area, 0xcc, body_area_size);
*/
  }
  sentry_shm_info->info.last_update = -1;
}

void
show_sentry_shm()
{
  /* no-op */
}

