/*
 * Sentry Video System
 *   is video server with QoS control from clients' requests.
 *
 *      original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *      modified by k-chinen@is.aist-nara.ac.jp, 1996
 *      modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *      modified by atsush-m@is.aist-nara.ac.jp, 2002
 *  
 *  $Id: $
 */

/*
 * send_image.c - video handling procedures.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <errno.h>

#include "alist.h"
#include "sentry.h"
#include "server.h"


extern fd_set writemask;
extern int maxwritefds;

extern _req_val req_val;

extern int multicast_clients;
extern int unicast_clients;

int   frameSize;
int   xsize, ysize;
int   capture_xsize, capture_ysize;
int   xl_offset=0, xr_offset=0, wide_offset;
int   yt_offset=0, yb_offset=0;
static unsigned char *dataPtr;
static unsigned char *new_dataPtr;

void show_stat();

void
send_image(int signo)
{
    static int cap_count=0;
    static long sum_takes=0;
    sentry_protocol3_image_hdr *hdr;
    struct timeval waittime;
    struct timeval cap_begin, cap_end;
    int new_xsize, new_ysize;
    int fd;
    int datalen, rest, offset, send_datalen, send_copunt;
    int trans_corder = SEN_F_NONE;
    int nwrite;
    unsigned char *sendptr;
    int trans_format;
    int takes;
    int ic;
    int delta;

#ifdef USE_IOVEC
    sentry_protocol3_image_hdr header;
    struct iovec iov[2];
#else
    char packet[8192];
#endif /* USE_IOVEC */
#ifdef ALIVE
    extern void set_alive_mark();
#endif /* ALIVE */
    extern void set_rate();



    if (unicast_clients <= 0 && multicast_clients <= 0) {
        signal(SIGALRM, send_image);
        return;
    }
    signal(SIGALRM, SIG_IGN);

    gettimeofday(&cap_begin, NULL);
    dataPtr = (char *)capture_frame();
    gettimeofday(&cap_end, NULL);
    cap_count++;

    if(debug>=100) {
        fprintf(stderr, "got image; ptr %p, len %d, format '%s'\n",
            dataPtr, frameSize, assoc_v(color_alist, capture_corder));
    }

    takes = SUB_TIME(cap_end, cap_begin);
    sum_takes += takes;

    if(debug>=5) {
        fprintf(stdout, "take %3d ms\n", takes);
    }
    if(debug) {
        if(cap_count%100==0) {
            fprintf(stdout, "sum takes %5ld ms, ave %.2f ms\n",
                sum_takes, ((double)sum_takes)/cap_count);

            sum_takes = 0;
            cap_count = 0;
        }
    }

    if(dataPtr==NULL) {
        fprintf(stderr, "ERROR: dataPtr %p, len %d\n",
            dataPtr, frameSize);
    }

#ifdef ALIVE
  (void)set_alive_mark();
#endif 

    /*
     * clip image for encode from captured image
     */
    if (xl_offset > 0 || xr_offset > 0 || wide_offset > 0 || yt_offset > 0) {
        if (clip_image(&dataPtr, &frameSize) < 0) {
          fprintf(stderr, "send_image: cannot clip image.\n");
        }
    }

    if(debug>=100) {
        fprintf(stderr, "b1 dataPtr %p, len %d, %dx%d\n",
            dataPtr, frameSize, xsize, ysize);
    }

    /*
     * convert video format
     */

    if (debug>=100) {
        fprintf(stderr, "send_image: req_val.attr.val[QOS_FORMAT=%d] = %d\n",
            QOS_FORMAT, req_val.attr.val[QOS_FORMAT]);
        fprintf(stderr, "  * requested format '%s'\n", 
            assoc_v(format_alist, req_val.attr.val[QOS_FORMAT]));
    }

    switch(req_val.attr.val[QOS_FORMAT]) {
    case SEN_F_NONE:
        new_dataPtr = dataPtr;
        new_xsize = xsize;
        new_ysize = ysize;
        datalen = frameSize;
/*
        trans_format = SEN_F_NONE;
*/
        trans_format = capture_format;
        trans_corder = capture_corder;
        break;

    case SEN_F_JPEG:
        if(debug>=200) {
            fprintf(stderr, " jpeg: before data %p, len %d, size %dx%d\n",
                dataPtr, frameSize, xsize, ysize);
        }

#if 0
    if (debug && frameSize>0) {
      static int bc=0;
      char filename[256];
      extern int dump_rgb_to_file(char *filename, char *image, int size,int);
      sprintf(filename, "s-r-sample%d.ppm", bc);
        dump_rgb_to_file(filename, dataPtr, xsize, ysize);
      bc = (bc+1) % 7;
    }
#endif
        datalen = convert_jpeg(dataPtr, frameSize, xsize, ysize,
                        &new_xsize, &new_ysize, &new_dataPtr);
        trans_format = SEN_F_JPEG;

        if(debug>=200) {
            fprintf(stderr, " jpeg: after  data %p, len %d, size %dx%d\n",
                new_dataPtr, datalen, new_xsize, new_ysize);
        }

#if 0
    if (debug && datalen>0) {
      static int bc=0;
      char filename[256];
      extern int dump_to_file(char *filename, char *image, int size);
      sprintf(filename, "s-o-sample%d.jpg", bc);
      dump_to_file(filename, new_dataPtr, datalen);
      bc = (bc+1) % 7;
    }
#endif
        break;

    case SEN_F_RGB:
        new_dataPtr = dataPtr;
        new_xsize = xsize;
        new_ysize = ysize;
        datalen = frameSize;
        trans_format = SEN_F_RGB;
        trans_corder = capture_corder;
        break;

    default:
        datalen = convert_none(dataPtr, frameSize, xsize, ysize, &new_xsize, &new_ysize, &new_dataPtr);
        trans_format = SEN_F_NONE;
        break;
    }
    if(debug>=100) {
        fprintf(stderr, "  in  dataPtr %p, len %6d, %dx%d\n",
            dataPtr, frameSize, xsize, ysize);
    }
    if(debug>=100) {
        fprintf(stderr, "  out dataPtr %p, len %6d, %dx%d\n",
            new_dataPtr, datalen, new_xsize, new_ysize);
    }
    if(debug>=100) {
        fprintf(stderr, "  e.g., 4bytes/pixel %d, 3bytes/pixel %d\n",
            new_xsize*new_ysize*4, new_xsize*new_ysize*3);
    }

    if(debug>=100) {
        fprintf(stderr, "  * encode format type '%s'\n",
            assoc_v(format_alist, trans_format));
    }
  
  /* send image */
#ifdef USE_IOVEC
    hdr = &header;
  
    hdr->hd_ver = SENTRY_PROTOCOL_VERSION;
    hdr->hd_ts[0] = htonl(cap_end.tv_sec);
    hdr->hd_ts[1] = htonl(cap_end.tv_usec);
  
    hdr->format = htonl(trans_format);
    hdr->size   = htonl(datalen);
    hdr->width  = htonl(new_xsize);
    hdr->height = htonl(new_ysize);
#if 0
    hdr->color  = htonl(req_val.attr.val[QOS_COLOR]);
#endif
    hdr->color  = htonl(trans_corder);

#if 0
fprintf(stderr, "color order %s(%d)\n",
    assoc_v(color_alist, trans_corder), trans_corder);
#endif
  
    /*
    ** iov
    **  #0  header
    **  #1  body
    */
    iov[0].iov_base = (char *)hdr;
    iov[0].iov_len = sizeof(header);
  
    /* wait a tick. I don't know why :-p */
#ifdef USE_WAITTIME
    waittime.tv_sec = 0;
    waittime.tv_usec = 30000;   /* 30 ms */
#endif
  
    /* 
     * send packets with unicast
     *  (writemask check)
     */
    for(fd=0; fd<=maxwritefds; fd++) {
        if (!FD_ISSET(fd, &writemask))
            continue;
#ifdef SUPPORT_MULTIRATE
        /* not implemented yet */
#endif
    
        offset = 0;
        rest = datalen;
        sendptr = new_dataPtr;
        /* send_count = 0; */

        if(debug>=100) {
            fprintf(stderr, "offset %d, rest %d, sendptr %p\n",
                offset, rest, sendptr);
            fprintf(stderr, " %d x %d + %d\n",
                rest/SENTRY_MAX_DATALEN, SENTRY_MAX_DATALEN,
                rest%SENTRY_MAX_DATALEN);
        }
    
#if 0
        if (debug) {
            fprintf(stderr, "send_image: sentptr# %x\n", sendptr);
        }
#endif

    
        ic = 0;
        while (rest > 0) {
            hdr->offset = htonl(offset);
      
#if 0
            iov[1].iov_base = sendptr;
#endif
#if 0
            iov[1].iov_base = &sendptr[offset];
#endif
            iov[1].iov_base = &sendptr[SENTRY_MAX_DATALEN*ic];

            if (rest >= SENTRY_MAX_DATALEN)
                iov[1].iov_len = SENTRY_MAX_DATALEN;
            else 
                iov[1].iov_len = rest;

  if(debug>=500) {
    printf("  %d: rest %d\n", ic, rest);
    printf("  iov[0] %#10x %4d\n", iov[0].iov_base, iov[0].iov_len);
    printf("  iov[1] %#10x %4d\n", iov[1].iov_base, iov[1].iov_len);
/*
    printf("    %d %x\n", sendptr[offset], sendptr[offset]);
*/
    delta = (unsigned char*)iov[1].iov_base - new_dataPtr;
    printf("    delta %5d = %2d x %d + %4d\n",
        delta, delta/SENTRY_MAX_DATALEN, SENTRY_MAX_DATALEN,
        delta%SENTRY_MAX_DATALEN);
  }
          
try_write2:
            nwrite = writev(fd, iov, 2);
            if (nwrite < 0) {
                perror("send_image: writev");
            
                if (errno==EINTR)
                    goto try_write2;
                clear_client(find_client_by_datafd(fd));
                goto fail2;
            } /* have finished sending packets */
          
            send_datalen = nwrite - sizeof(header);
/*
            sendptr = &sendptr[send_datalen];
*/
            offset += send_datalen;
            rest -= send_datalen;

  if(debug>=500) {
    printf("     nwrite %d, send_datalen %d\n", nwrite, send_datalen);
  }
      
#if 0
            if (debug) {
                if (offset == send_datalen || send_datalen < 8156) 
                    fprintf(stderr,
"send_image: ts# %d.%06d, rest# %d, send_datalen# %d, nwrite# %d\n",
                ntohl(hdr->hd_ts[0]), ntohl(hdr->hd_ts[1]),
                rest, send_datalen, nwrite);
            }
#endif

            ic++;
        } /* while for write-loop */
    
fail2:;
    }/* for-loop */
  
  /*
   * send packets with (udp) multicast
   */
  if (multicast_clients > 0) {
    hdr->offset = 0;
    rest = datalen;
    
    sendptr = new_dataPtr;
    /* send_count = 0; */
    
    while (rest > 0) {
      iov[1].iov_base = sendptr;
      if (rest >= SENTRY_MAX_DATALEN)
        iov[1].iov_len = SENTRY_MAX_DATALEN;
      else 
        iov[1].iov_len = rest;
      
try_write:
      nwrite = writev(fd, iov, 2);
      if (nwrite < 0) {
        fprintf(stderr, "send_image: writev: %s\n", sys_errlist[errno]);

        if (errno==EINTR)
          goto try_write;

        clear_client(find_client_by_datafd(fd));
        goto fail;
      } /* have finished sending packets */
      
      send_datalen = nwrite - sizeof(header);
      hdr->offset += send_datalen;
      sendptr = &sendptr[send_datalen];
      rest -= send_datalen;
    } /* while for write-loop */
    
fail:;
  }
  
#else /* NOT IOVEC */
  fprintf(stderr, "Now, only USE_IOVEC version is supported.\n");
  exit (1);
#endif /* USE_IOVEC */

  release_frame();
  
  signal(SIGALRM, send_image);
  /*
   * for statistics information
   */
  show_stat();
}

/*
 * clip image
 */
int 
clip_image(char **data, int *size)
{
  int color_depth, i, n, x_offset;
  char *ptr, tmp;

  PASS;
  
  switch(req_val.attr.val[QOS_COLOR]) {
  case FULL_COLOR:
#ifdef IRIX 
    color_depth = 4;
#else
    color_depth = 3;
#endif
    break;
  case MONOCHROME:
  default:
    color_depth = 1;
    break;
  }

#if 0
  fprintf(stderr, "color_depth: %d\n", color_depth);
#endif

  (*size) = xsize * ysize * color_depth;

  if(wide_offset > 0)
    *data = &(*data)[capture_xsize*wide_offset*color_depth];

  if(yt_offset > 0)
    *data = &(*data)[capture_xsize*yt_offset*color_depth];

#ifndef IRIX
  if(xl_offset > 0 || xr_offset > 0){
    x_offset = xl_offset + xr_offset;
    *data = &(*data)[xl_offset*color_depth];
    for(i=0, ptr=&(*data)[xsize*color_depth]; i<ysize; i++){
      for(n=0; n<xsize*color_depth; n++){
        ptr[0] = ptr[x_offset*color_depth];
        ptr++;
      }
      x_offset = x_offset + xl_offset + xr_offset;
    }
  }
#endif
  return 0;
}

#ifdef ALIVE
/*
 * plot alive mark in top line
 */
void
set_alive_mark(void)
{
  int i, color_depth;
  static int alive_pos = 0;

  switch(req_val.attr.val[QOS_COLOR]) {
  case FULL_COLOR:
#ifdef IRIX
    color_depth = 4;
#else
    color_depth = 3;
#endif
    break;
  case MONOCHROME:
  default:
    color_depth = 1;
    break;
  }

  for(i=0; i<color_depth*2; i++) {
    dataPtr[alive_pos+i] = 0xff;
  }

  if (alive_pos+i-1 < xsize*color_depth)
    alive_pos += i;
  else
    alive_pos = 0;
}
#endif /* ALIVE */

/*
 * show send_image statistics information 
 */
void show_stat() 
{
  static int FrameCount = 0;
  static int TotalCount = -1;
  static struct timeval start_tv = {0, 0};
  struct timeval tv;
  
  if (start_tv.tv_sec == 0) 
    gettimeofday(&start_tv, NULL);
  gettimeofday(&tv, NULL);

  if (++TotalCount >= 1000) {
    if (tv.tv_usec -= start_tv.tv_usec) {
      tv.tv_sec -= 1;
      tv.tv_usec += 1000000;
    }
    tv.tv_sec -= start_tv.tv_sec;

    fprintf(stdout, "Total: frame# %d, uptime: %d.%06d s, rate# %.2f fps\n",
            FrameCount, tv.tv_sec, tv.tv_usec, 
            (((double)FrameCount)*1000000 / (tv.tv_sec*1000000+tv.tv_usec)));
    TotalCount = 0;
  }
  FrameCount ++;
}
