/*
 * sentry.c - image viewer for sentry protocol
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/cursorfont.h>

#include <X11/Xmu/WinUtil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Scrollbar.h>

/* private TrueColor Widget */
#include "Lww.h"

#include "alist.h"
#include "sentry.h"
#include "img.h"
#include "pool.h"
#include "client.h"



extern  void request_call();

extern u_char client_type; 
extern int server_zoom;
extern int server_size;
extern int server_color;
extern int server_format; 
extern int server_rate;

static int show_help=0;


typedef struct {
    String server_name;
    String server_port;
    String data_port;
    String req_rate;
    String req_type;
    String req_size;
    String req_zoom;
    String req_color;
    String req_format;
    String hivision_flag;
    String voffset;
    /* */
    String shm_key;
    /* */
    String compare_func;
    /* */
    String debug_level;
    String show_help;
} sentry_resource_t;

sentry_resource_t sentry_resource;

#define Offset(field) (XtOffsetOf(sentry_resource_t, field))
static XtResource resources[] = {
  {"servername", "ServerName",  XtRString,  sizeof(String),
                    Offset(server_name), XtRString,  NULL},
  {"serverport", "ServerPort",  XtRString,  sizeof(String),
                    Offset(server_port), XtRString,  NULL},
  {"dataport",   "DataPort",  XtRString,  sizeof(String),
                    Offset(data_port), XtRString,  NULL},
  {"reqsize",   "ReqSize",  XtRString,  sizeof(String),
                    Offset(req_size), XtRString,  NULL},
  {"reqrate",   "ReqRate",  XtRString,  sizeof(String),
                    Offset(req_rate), XtRString,  NULL},
  {"reqformat", "ReqFormat",  XtRString,  sizeof(String),
                    Offset(req_format), XtRString,  NULL},
  {"reqtype",   "ReqType",  XtRString,  sizeof(String),
                    Offset(req_type), XtRString,  NULL},

  {"shmkey",    "ShmKey",  XtRString,  sizeof(String),
                    Offset(shm_key), XtRString,  NULL},

  {"comp",      "comp",  XtRString,  sizeof(String),
                    Offset(compare_func), XtRString,  NULL},

  {"debug",     "Debug",  XtRString,  sizeof(String),
                    Offset(debug_level), XtRString,  NULL},
  {"help",      "Help",  XtRString,  sizeof(String),
                    Offset(show_help), XtRString,  NULL},
};




String fallback_resources[] = {
    "*reqsize:      160x120",
    "*reqrate:      200",
    "*reqformat:    JPEG",
    "*servername:   localhost",
    "*serverport:   7001",
    "*dataport:     7000",

    "*shmkey:       -1",

    "*comp:         none",

    "*debug:        0",

    "*help:         0",

    "*foreground:   black",
    "*background:   white",
    NULL};

static XrmOptionDescRec options[] = {
{"-fg",         "*Foreground",  XrmoptionSepArg,    NULL},
{"-bg",         "*Background",  XrmoptionSepArg,    NULL},
{"-servername", "*servername",  XrmoptionSepArg,    NULL},
{"-serverport", "*serverport",  XrmoptionSepArg,    NULL},
{"-dataport",   "*dataport",    XrmoptionSepArg,    NULL},

{"-shmkey",     "*shmkey",      XrmoptionSepArg,    NULL},
{"-debug",      "*debug",       XrmoptionSepArg,    NULL},
{"-rate",       "*reqrate",     XrmoptionSepArg,    NULL},
{"-format",     "*reqformat",   XrmoptionSepArg,    NULL},
{"-size",       "*reqsize",     XrmoptionSepArg,    NULL},

{"-comp",       "*comp",        XrmoptionSepArg,    NULL},

{"-help",       "*help",        XrmoptionNoArg,     "1"},
};

Widget   toplevel;
XtAppContext app_context;
XtInputId   iid, ciid;

XtActionsRec main_action_list[] = {
    NULL,
};


Visual   *xp_slide_visual,   *xp_foil_visual;
int      xp_slide_depth,     xp_foil_depth;
Colormap xp_slide_cmap,      xp_foil_cmap;
int      xp_use_special_cmap;

String fvisual, svisual;



Widget pwin;
Widget pshell;
Widget pbase;
Widget plabel;
GC  pwin_gc;

typedef struct {
    XtIntervalId    id;
    time_t          interval;
} xp_periodical_rec;

xp_periodical_rec win1;


void KickWin(XtPointer client_data, XtIntervalId *iid);


extern sentry_shmrec   *sentry_shm_chunk;


show_chunk()
{
    int i;
    int pos;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
    }
}

int 
find_pos(time_t reftime)
{
    int i;
    int pos;
    int ret;

    ret = -1;

    for(i=0;i<sentry_shm_chunk_radix;i++) {
        pos = i+1;
#if 0
        printf(" %3d: %3d %12d %4d\n", i, pos,
            sentry_shm_chunk[pos].entry.header.recv_sec,
            sentry_shm_chunk[pos].entry.header.length);
#endif
        if(sentry_shm_chunk[pos].entry.header.recv_sec==reftime) {
            if(debug>=10) {
                printf(
" %3d: %3d time issue %d.%06d recv %d.%06d len %4d (%dx%d) pos %p [%d]\n",
                    i, pos,
                    sentry_shm_chunk[pos].entry.header.send_sec,
                    sentry_shm_chunk[pos].entry.header.send_usec,
                    sentry_shm_chunk[pos].entry.header.recv_sec,
                    sentry_shm_chunk[pos].entry.header.recv_usec,
                    sentry_shm_chunk[pos].entry.header.length,
                    sentry_shm_chunk[pos].entry.header.width,
                    sentry_shm_chunk[pos].entry.header.height,
                    sentry_shm_chunk[pos].entry.header.body,
                    sentry_shm_chunk[pos].entry.header.body_len
                    );
            }
            ret = pos;
        }
    }

    return ret;
}


SenDrawImage2(Widget w, sentry_imgrec *inimg, sentry_imgrec *outimg)
{
    int wid, ht;
    int xwid, xht;
    unsigned char *raw_data;
    unsigned char *data_rgb=NULL, *data_bgrx=NULL;
    int raw_len;
    int raw_format, raw_color;
    int chk;
    int cur_format=SEN_F_NONE;
    int cur_corder=SEN_C_NONE;
    int success=0;
    int outlen=-1;

    if(inimg==NULL)
        return;

#if 0
    if(debug>=10) {
        printf("SenDrawImage: output format '%s', color order '%s'\n",
            assoc_v(format_alist, output_format),   
            assoc_v(color_alist, output_corder));
    }
#endif

    raw_len     = inimg->length;
    raw_data    = (unsigned char*)inimg->body;
    raw_format  = inimg->format;
    raw_color   = inimg->corder;
    wid = inimg->width;
    ht  = inimg->height;

    data_rgb = (unsigned char*)malloc(wid*ht*3);
    memset(data_rgb, 0xff, wid*ht*3);
    data_bgrx = (unsigned char*)malloc(wid*ht*4);
    memset(data_bgrx, 0xff, wid*ht*4);

#if 0
    if(debug>=10) {
        printf("SenDrawImage: %p %d '%s''%s' %dx%d, 3B/p=%d 4B/p=%d\n",
            raw_data, raw_len, assoc_v(format_alist, raw_format),
            assoc_v(color_alist, raw_color), wid, ht, wid*ht*3, wid*ht*4);
    }
#endif
    if(debug>=10) {
        printf(
"draw: got format %s, color %s, ptr %p, size %dx%d, len %d\n\
      out format %s, color %s\n",
        assoc_v(format_alist, raw_format), assoc_v(color_alist, raw_color), 
        raw_data, wid, ht, raw_len,
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder));
    }

    switch(raw_format) {
    case SEN_F_JPEG:
        chk = decode_jpeg(raw_data, raw_len, data_rgb, &xwid, &xht);
        if(xwid!=wid && xht!=ht) {
            fprintf(stderr,
                "SenDrawImage: wid %d != xwid %d, ht %d != xht %d\n",
                wid, xwid, ht, xht);
        }
        cur_format = SEN_F_RGB;
        cur_corder = SEN_C_RGB;
        if(debug>=20){
            printf("  convert to '%s' (wid %d, ht %d\n",
                assoc_v(color_alist, cur_corder), xwid, xht);
        }
        break;

    case SEN_F_NONE: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        break;

    case SEN_F_RGB: /* XXX */
        memcpy(data_rgb, raw_data, raw_len);
        cur_format = raw_format; /* SEN_F_RGB */
        cur_corder = raw_color;
        break;
    }

    if(debug>=10) {
        printf(
"draw: cur format %s, color %s\n\
      out format %s, color %s\n",
        assoc_v(format_alist, cur_format), assoc_v(color_alist, cur_corder),
        assoc_v(format_alist, output_format),
        assoc_v(color_alist, output_corder)
        );
    }
    if(cur_format != output_format) {
        if(debug>=1) {
            fprintf(stderr, "image format miss-match cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
        goto put_fail;
    }

    if(output_format==SEN_F_RGB) {
        if(output_corder==SEN_C_XBGR) {
            convert_RGB2XBGR(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), wid, ht, data_bgrx);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_XRGB) {
/*
            convert_RGB2XRGB(data_bgrx, data_rgb, wid * ht);
*/
            convert_RGB2BGRX(data_bgrx, data_rgb, wid * ht);

            outlen = wid*ht*4;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), wid, ht, data_bgrx);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_bgrx, outlen);
        }
        else
        if(output_corder==SEN_C_RGB) {
            outlen = wid*ht*3;
            put_image(XtDisplay(w), xp_foil_visual,
                xp_foil_depth, pwin_gc, XtWindow(w), wid, ht, data_rgb);
            success = 1;

            copyattr_img(outimg, inimg);
            outimg->format = output_format;
            outimg->corder = output_corder;
            outimg->length = outlen;
            memcpy(outimg->body, data_rgb, outlen);
        }
    }

put_fail:
    if(!success) {
        if(debug>=1) {
            fprintf(stderr, "fail some reason, format cur=%s out=%s\n",
                assoc_v(format_alist, cur_format),
                assoc_v(format_alist, output_format));
        }
    }
    else {
#if 0
        copyattr_img(outimg, inimg);
        memcpy(outimg->body, inimg->body, outlen);
#endif
    }

out:
    if(data_rgb!=NULL)
        free(data_rgb);
    if(data_bgrx!=NULL) 
        free(data_bgrx);
}


static int call_count=0, lost_count=0, got_count=0, c_got_count=0;
static int refresh_flag = 1;


RefreshWins(sentry_imgrec *bank)
{
    Dimension tw, th, tx, ty;
    Dimension bw, bh, bx, by;
    Dimension nw, nh;

    XtVaGetValues(plabel,
         XtNx, &tx, XtNy, &ty, XtNwidth, &tw, XtNheight, &th, NULL);
    XtVaGetValues(pwin,
         XtNx, &bx, XtNy, &by, XtNwidth, &bw, XtNheight, &bh, NULL);
#if 0
    printf("1 %ux%u+%u+%u %ux%u+%u+%u\n",
        tw, th, tx, ty, bw, bh, bx, by);
#endif

    XtVaSetValues(plabel,
        XtNwidth, bank->width,
        XtNlabel, server_host,
        NULL);

    XtVaGetValues(plabel, XtNheight, &th, NULL);
    XtVaGetValues(pwin, XtNheight, &bh, NULL);
    XtVaGetValues(plabel, XtNy, &ty, NULL);
    XtVaGetValues(pwin, XtNy, &by, NULL);
#if 0
    printf("2 th %d, bh %d, ty %d, by %d\n", th, bh, ty, by);
#endif


    XtVaSetValues(pwin,
        XtNwidth, bank->width,
        XtNheight, bank->height,
        NULL);

    nw = bank->width+10;
/*
    nh = bank->height+th+10;
*/
    nh = MAX(by,24) + bank->height+10;

#if 0
    printf("nw %d, nh %d\n", nw, nh);
#endif

    XtVaSetValues(pshell,
        XtNwidth, nw, XtNheight, nh,
/*
*/
        NULL);
    XtVaSetValues(pbase,
/*
        XtNwidth, bank->width,
        XtNheight, bank->height,
*/
/*
        XtNwidth, nw, XtNheight, nh,
*/
        NULL);

    XtVaGetValues(plabel, XtNheight, &th, NULL);
    XtVaGetValues(pwin, XtNheight, &bh, NULL);
#if 0
    printf("3 th %d, bh %d\n", th, bh);
#endif

    fflush(stdout);
    fflush(stderr);

    refresh_flag = 0;
}




#define TMP_IMGBUF_LEN  (100*1024)

static sentry_shmrec bk_info;
static sentry_imgrec bk_inframe;
static unsigned char *inimgbuf=NULL;
static sentry_imgrec bk_drframe[2];
static unsigned char *drimgbuf0=NULL;
static unsigned char *drimgbuf1=NULL;



extern apair_t buflen_xrgb_alist;


alloc_inimgbufs()
{
    inimgbuf = (unsigned char*)malloc(TMP_IMGBUF_LEN);
    bk_inframe.body = inimgbuf;
    bk_inframe.body_len = TMP_IMGBUF_LEN;
}

alloc_drimgbufs()
{
    int buflen;

    if(output_format!=SEN_F_RGB) {
        fprintf(stderr, "no idea for no RGB device\n");
        exit(1);
    }

#if 0
    if(output_corder==SEN_C_RGB||
        output_corder==SEN_C_BGR) {
        buflen = TMP_IMGBUF_LEN;
    }
    else {
        buflen = TMP_IMGBUF_LEN;
    }
#endif
    buflen = (int)assoc_p(&buflen_xrgb_alist, MAX_DISPLAY_SIZE);
 if(debug) {
    printf("buflen %d\n", buflen);
 }
    if(buflen==0) {
        buflen = TMP_IMGBUF_LEN;
    }

    drimgbuf0 = (unsigned char*)malloc(buflen);
    bk_drframe[0].body = drimgbuf0;
    bk_drframe[0].body_len = buflen;

    drimgbuf1 = (unsigned char*)malloc(buflen);
    bk_drframe[1].body = drimgbuf1;
    bk_drframe[1].body_len = buflen;
}


int diff_capacity = 0x20;
int cmp_type = SEN_M_NONE;

void
x_get_data(caddr_t client_data, int *fd, XtInputId *id)
{
    static struct timeval lasttime={-1,0};
    struct timeval curtime;
    time_t now;
    int chk;
    sentry_shmrec *info;
    sentry_imgrec *inframe;
    static sentry_imgrec *cur=NULL, *last=NULL;
    char *tmp_buff;
    int pos;
    int diff;
    double rate;
    int  drfno;
    char *drframe;
    int get_data3(int datafd, char **buff,
        sentry_shmrec *info, sentry_imgrec *dest);

    gettimeofday(&curtime, NULL);
    now = curtime.tv_sec;

    if(use_shm_flag) {
        info = &sentry_shm_chunk[0];
        pos  = 1+now%sentry_shm_chunk_radix;
        inframe = &(sentry_shm_chunk[pos].entry.header);
    }
    else {
        info = &bk_info;

        if(c_got_count==0) {
            alloc_inimgbufs();
            alloc_drimgbufs();
        }

        inframe = &bk_inframe;
    }

 if(debug>=100) {
    fprintf(stderr, "inframe %p\n", inframe);
 }

    chk = get_data3(*fd, &tmp_buff, info, inframe);
    if(chk) {
        return;
    }

    if(got_count%report_interval==0) {
        syslog(LOG_NOTICE, "sentryc key=%d; got %d data\n",
            sentry_shmkey, got_count);
    }
    if(refresh_flag){
        RefreshWins(inframe);
    }

    drfno = c_got_count%2;
    cur = &bk_drframe[drfno];

 if(debug>=100) {
    fprintf(stderr, "drfno %d, bk_drframe[%d] %p, cur %p\n",
        drfno, drfno, &bk_drframe[drfno], cur);
 }

    SenDrawImage2(pwin, inframe, cur);

    if(last) {
        if(debug) {
            show_imgattr(stderr, cur);
            show_imgattr(stderr, last);
        }

        chk = img_cmp(cur, last, cmp_type, (void*)&diff_capacity);
        if(debug) {
            fprintf(stderr, "DIFF %d\a\a\a\n", chk);
        }
    }


    got_count++;
    c_got_count++;

    if(curtime.tv_sec-lasttime.tv_sec>report_interval) {
        if(lasttime.tv_sec<=0) {
        }
        else {
            diff = SUB_TIME(curtime, lasttime);
            rate = (double)got_count*1000.0/diff;
            fprintf(stderr, "##### rate %5.2f\n", rate);
        }
        lasttime = curtime;
        got_count = 0;
    }

    last = cur;

    return;
}

void KickWin(XtPointer client_data, XtIntervalId *iid)
{
    struct timeval now;
    xp_periodical_rec *prof;
    int chk;
    int pos;
    sentry_imgrec *inframe;
    int  drfno;
    char *drframe;
    static sentry_imgrec *cur=NULL, *last=NULL;

    gettimeofday(&now, NULL);
    if(debug>=10) {
        printf("KickWin: %d.%06d %p %p\n",
            now.tv_sec, now.tv_usec, client_data, iid);
    }
    prof = (xp_periodical_rec*) client_data;

/*
    show_chunk();
*/
    pos = find_pos(now.tv_sec);
    if(debug>=20) {
        printf("pos %d\n", pos);
    }
    if(pos>=0) {
#if 0
        if(got_count<=0) {
            RefreshWins(&sentry_shm_chunk[pos]);
        }
#endif
        if(c_got_count==0) {
            alloc_drimgbufs();
        }
        inframe = &sentry_shm_chunk[pos];

        if(refresh_flag){
            RefreshWins(&sentry_shm_chunk[pos].entry.header);
        }

        drfno = c_got_count%2;
        cur = &bk_drframe[drfno];

/*
        SenDrawImage(pwin, &sentry_shm_chunk[pos]);
*/
        SenDrawImage2(pwin, inframe, cur);

        if(last) {
            if(debug) {
                show_imgattr(stderr, cur);
                show_imgattr(stderr, last);
            }

            chk = img_cmp(cur, last, SEN_M_9POINTS, (void*)&diff_capacity);
            if(debug) {
                fprintf(stderr, "DIFF %d\n", chk);
            }
        }

        got_count++;

        last = cur;
    }
    else {
        lost_count++;
    }

    if(debug<10) {
        printf(".");
        fflush(stdout);
    }


    prof->id = XtAppAddTimeOut(app_context,
        prof->interval, KickWin, prof);

    call_count++;

    if(call_count%30==0) {
        fprintf(stderr, "##### total %d, got %d, lost %d, err%% %.2f\n",
            call_count, got_count, lost_count,
            (double)lost_count*100.0/call_count);
        fflush(stderr);
    }


    return;
}



void
DoExpose(Widget w, XEvent *ev, String *s, Cardinal *c)
{
/*
    printf("DoExpose\n"); fflush(stdout);
*/
}

void
DoResize(Widget w, XEvent *ev, String *s, Cardinal *c)
{
    refresh_flag = 1;
}

static String pwin_trans="#override\n\
    <Expose>:           pwin-expose()\n\
    <Configure>:        pwin-resize()\n\
    <Key>:              pwin-expose()\n\
";

static XtActionsRec pwin_action_list[] = {
    {"pwin-expose",     DoExpose},
    {"pwin-resize",     DoResize},
};

void
usage()
{
        printf("sentry - image viewer of sentry system\n\
usage: sentry [options]\n\
\n\
option:\n\
    -help\n\
    -servername host    default host is localhost\n\
    -serverport port    \n\
    -dataport port      \n\
    -reqrate num        frame rate within per 100sec\n\
    -format format      frame format (e.g., NONE, JPEG, PNG, etc.)\n\
    -shmkey key         \n\
    -debug num          debug leve\n\
\n\
note:\n\
  * you should change dataport when you start multipul program\n\
  * frame rate does not changed by this program, if you access\n\
    with shared memory\n\
\n\
example:\n\
  %% sentry -severname camera1 -dataport 723\n\
  %% sentry -shmkey 129\n\
");
}




fd_set  InputFdMask;


int
main(int argc, char **argv)
{
    int skey;
    XGCValues values;
    int valuemask;
    int tv;


    toplevel = XtVaAppInitialize(
        &app_context,                   /* Application context */
        "Xp",                           /* Application class */
        options, XtNumber(options),     /* option list */
        &argc, argv,                    /* command line args */
        fallback_resources,             /* for missing app-defaults file */
        XtNtitle,       "XP",
        XtNiconName,    "XP",
        NULL);                          /* terminate varargs list */

    XtVaGetApplicationResources(toplevel,
        (XtPointer) &sentry_resource, resources, XtNumber(resources), NULL);

    XtAppAddActions(app_context, main_action_list, XtNumber(main_action_list));



    if(getenv("SENTRY_SERVER")) {
        strncpy(server_host, getenv("SENTRY_SERVER"), 127);
    }
    else {
        strncpy(server_host, sentry_resource.server_name, 127);
    }
    server_host[127]='\0';

    server_port = atoi(sentry_resource.server_port);
    data_port   = atoi(sentry_resource.data_port);
    server_rate = atoi(sentry_resource.req_rate);
    debug = atoi(sentry_resource.debug_level);
    show_help = atoi(sentry_resource.show_help);
    if(atoi(sentry_resource.shm_key)<=0)
        use_shm_flag = 0;
    else {
        use_shm_flag = 1;
        skey = atoi(sentry_resource.shm_key);
    }

    tv = assoc_n(matching_type_alist, sentry_resource.compare_func);
    if(debug) {
        fprintf(stderr, "compare type %d %x '%s' <= '%s'\n",
            tv, tv, assoc_v(matching_type_alist, tv),
            sentry_resource.compare_func);
    }
    cmp_type = tv;

    if(debug) {
        fprintf(stderr, "severname '%s'\n", sentry_resource.server_name);
    }

    tv = assoc_n(format_alist, sentry_resource.req_format);
    server_format = tv;

    tv = assoc_n(size_alist, sentry_resource.req_size);
    server_size = tv;

    if(debug>=10) {
    fprintf(stderr, "format %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_format, assoc_v(format_alist, tv));
    fprintf(stderr, "size %d, '%s' => '%s'\n",    
        tv, sentry_resource.req_size, assoc_v(size_alist, tv));
    }


    if(show_help) {
        usage();
        exit(0);
    }

    fflush(stdout);
    fflush(stderr);
    SetupVisual(XtDisplay(toplevel), fvisual, svisual);
    fflush(stdout);
    fflush(stderr);

    output_format = SEN_F_RGB;
#if 0
    /* corder is solved by SetupVisual() */
    output_corder = SEN_F_RGB;
#endif


    pshell = XtVaAppCreateShell(
        "sentry", "Sentry",
        topLevelShellWidgetClass, XtDisplay(toplevel),
        XtNvisual, xp_foil_visual,
        XtNdepth, xp_foil_depth,
        XtNcolormap, xp_foil_cmap,
        NULL);

    pbase = XtVaCreateManagedWidget(
        "base", formWidgetClass, pshell,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
        NULL);
    plabel = XtVaCreateManagedWidget(
        "label", labelWidgetClass, pbase,
        XtNlabel, "boot...",
        XtNjustify, XtJustifyLeft,
/*
        XtNforeground, 0xff00ff,
        XtNbackground, 0x00ff00,
*/
/*
        XtNwidth, 120,
*/
        XtNresize,  False,
        NULL);

    pwin = XtVaCreateManagedWidget(
        "Pwin", lwwWidgetClass, pbase,
        XtNfromVert, plabel,
        XtNvisual,  xp_foil_visual,
        XtNdepth,   xp_foil_depth,
        XtNcolormap,xp_foil_cmap,
/*
        XtNwidth,  320, XtNheight, 240,
*/
        XtNwidth,  160, XtNheight, 120,
        NULL);

    XtRealizeWidget(pshell);

#if 0
    values.background =
        values.foreground = WhitePixel(XtDisplay(pwin),
                                DefaultScreen(XtDisplay(pwin)));
    valuemask = GCForeground | GCBackground;
    pwin_gc = XCreateGC(XtDisplay(pwin), XtWindow(pwin), valuemask, &values);
#endif
#if 1
    pwin_gc = XtGetGC(pwin, NULL, NULL);
#endif

    XtAppAddActions(app_context,
        pwin_action_list, XtNumber(pwin_action_list));
/*
    XtOverrideTranslations(pwin,    XtParseTranslationTable(pwin_trans));
*/
    XtOverrideTranslations(pbase,   XtParseTranslationTable(pwin_trans));
    XtOverrideTranslations(pshell,  XtParseTranslationTable(pwin_trans));



PASS;
    if(use_shm_flag) {
        init_shm(skey);
#if 0
        if(clear) {
            clear_shm();
        }
#endif
    }
    else {
        ctrlfd = build_ctrl_port(server_host, server_port);
        if (ctrlfd < 0) {
            printf("fail setup control fd with port# %d\n", server_port);
            exit (1);
        }

PASS;
        datafd = build_data_port(data_port);
        if (datafd < 0) {
            printf("fail setup data fd with port# %d\n", data_port);
            exit (1);
        }
    }
PASS;


    if(use_shm_flag) {
        fprintf(stderr, "sentry image viewer upon shared memory\n");

        win1.interval = 300; /* by ms */

        win1.id = XtAppAddTimeOut(app_context,
                   win1.interval, KickWin, &win1);
    }
    else {
        fprintf(stderr, "sentry image viewer with direct network access\n");

        init_ibuf();

        init_request();
        request_call();
        signal(SIGALRM, request_call);

PASS;

        FD_ZERO(&InputFdMask);
        FD_SET(ctrlfd, &InputFdMask);
        FD_SET(datafd, &InputFdMask);

        iid = XtAppAddInput(app_context, datafd,
                (caddr_t)XtInputReadMask, x_get_data, NULL);
#if 0
        ciid = XtAppAddInput(app_context, ctrlfd,
                (caddr_t)XtInputReadMask, retry_connect, NULL);
#endif

    }

PASS;

    XtAppMainLoop(app_context);

}



