/*
 * Sentry Video System
 *   is video server with QoS control from clients' requests.
 *
 *      original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *      modified by k-chinen@is.aist-nara.ac.jp, 1996
 *      modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *  
 *  $Id: $
 */

/*
 * sentry.h -- sentry video system general header
 */ 

#ifndef _SENTRY_H_
#define _SENTRY_H_

/*
#include <sys/time.h>
*/
#define SUB_TIME(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )

extern int debug;


/*
#include "alist.h"
*/


/*
 * address, port
 */
#define SENTRY_MULTICAST_ADDR   "224.0.0.1"
#define SENTRY_CTRL_PORT    7001
#define SENTRY_DATA_PORT    7000

/*
 * sentry current version
 */
#define SENTRY_PROTOCOL_VERSION      3


/*
 * sentry request type
 */
#define I_AM_ALIVE          (0x00)
#define OPEN_UDP_UNICAST    (0xf0)
#define OPEN_UDP_MULTICAST  (0xf1)
#define OPEN_TCP            (0xf2)
#define CLOSE               (0xff)

extern apair_t req_type_alist[];

#define QOS_REQUEST_SIZE    (0x01)
#define QOS_REQUEST_ZOOM    (0x02)
#define QOS_REQUEST_COLOR   (0x04)
#define QOS_REQUEST_FORMAT  (0x08)
#define QOS_REQUEST_RATE    (0x10)

extern apair_t req_attr_alist[];
extern apair_t req_desc_alist[];



/*
 * normal vision ratio (4:3)
 */
#define SIZE_UNKNOWN    (0)
#define SIZE_640x480    (1) /* VGA size */
#define SIZE_320x240    (2) /* quarter VGA size */
#define SIZE_212x160    (3) /* quarter CIF size */
#define SIZE_160x120    (4) 
/*
 * wide vision ratio (16:9)
 */
#define SIZE_640x360    (241)
#define SIZE_320x180    (242)
#define SIZE_212x120    (243)
#define SIZE_160x90     (244)

extern apair_t size_alist[];

/*
 * color 
 */
#define FULL_COLOR      (0x01)
#define MONOCHROME      (0x02)

/* color order */
#define SEN_C_NONE      (0x00)
#define SEN_C_GRAY      (0x02)
#define SEN_C_RGB       (0x01)
#define SEN_C_BGR       (0x04)
#define SEN_C_XRGB      (0x21)
#define SEN_C_RGBX      (0x41)
#define SEN_C_RGBA      (0x81)
#define SEN_C_XBGR      (0x24)
#define SEN_C_BGRX      (0x44)
#define SEN_C_BGRA      (0x84)

extern apair_t color_alist[];

/*
 * video format
 */
#define SEN_F_NONE      (0x00)  
#define SEN_F_JPEG      (0x01)
#define SEN_F_PNG       (0x02)
#define SEN_F_RGB       (0x41)

extern apair_t format_alist[];

/*
 * packet header 
 */
#if 0
typedef struct {
  u_char hd_ver;        /* sentry version */
  u_char type;          /* request type */
  short port;           /* data port# */

  int size;             /* frame size */
  int zoom;             /* zoom ratio */
  int color;            /* colot type */
  int format;           /* video format */
  int rate;             /* frame rate */
} sentry_protocol3_request_hdr;
#endif

typedef struct {
  uint8_t   hd_ver;     /* sentry version */
  uint8_t   type;       /* request type */
  uint16_t  port;       /* data port# */

  int32_t   size;       /* frame size */
  int32_t   zoom;       /* zoom ratio */
  int32_t   color;      /* colot type */
  int32_t   format;     /* video format */
  int32_t   rate;       /* frame rate */
} sentry_protocol3_request_hdr;

/*
 * sentry image header
 */

typedef struct {
  uint8_t   hd_ver;     /* sentry version */
  uint8_t   dummy[3];   /* padding */
  int32_t   offset;     /* image offset */
  union {
    int32_t ts[2];
    struct timeval timestamp; 
  } _hd_ts;             /* time stamp */
#define hd_ts      _hd_ts.ts 
#define hd_tstamp   _hd_ts.timestamp
  int32_t   format;     /* video format */
  int32_t   size;       /* image size (xsize * ysize * color_space) */
  int32_t   width;      /* image xsize */
  int32_t   height;     /* image ysize */
  int32_t   color;      /* color depth */
} sentry_protocol3_image_hdr;

#if 0
typedef struct {
#if 0
  struct rtp_hdr;
#endif
  u_char hd_ver;        /* sentry version */
  u_char dummy[3];      /* padding */
  int    offset;        /* image offset */
  union {
    long ts[2];
    struct timeval timestamp; 
  } _hd_ts;             /* time stamp */
#define hd_ts      _hd_ts.ts 
#define hd_tstamp   _hd_ts.timestamp
  int format;           /* video format */
  int size;             /* image size (xsize * ysize * color_space) */
  int width;            /* image xsize */
  int height;           /* image ysize */
  int color;            /* color depth */
} sentry_protocol3_image_hdr;
#endif


#define SENTRY_IMAGE_HEADER_LEN sizeof(sentry_protocol3_image_hdr)
#define SENTRY_MAX_DATALEN      (8192-SENTRY_IMAGE_HEADER_LEN)

#ifndef NOTRACE
#ifndef PASS
#define PASS  fprintf(stderr,"PASS %s:%d\n", __FILE__, __LINE__)
#endif
#else
#define PASS  (void)0
#endif





#endif /* _SENTRY_H_ */
