/*
 *
 * $Id$
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#ifdef TIME_WITH_SYS_TIME
 #include <sys/time.h>
 #include <time.h>
#else
 #ifdef HAVE_SYS_TIME_H
  #include <sys/time.h>
 #else
  #include <time.h>
 #endif
#endif
#include <errno.h>

#include "alist.h"
#include "img.h"
#include "fpair.h"
#include "pool.h"
#include "sentry.h"
#include "client.h"

/*
 * in file, images are recorded with network order
 * you have to convert them to host order.
 */

typedef struct {
    char magick[16];
    int32_t ver;
    int32_t len;
} token_t;


#define MAX_IMGSIZE     (128*1024)


sentry_imgrec img_pool[IMG_POOL_MAX];
int img_pool_count = 0;
int img_heap_use = 0;
fpair_t fpairs[IMG_POOL_MAX];


/*
 * fname    source file path
 * rc       record counter
 */

int
fpair_cmp(const void *xp, const void *yp)
{
    fpair_t *x, *y;
    x = (fpair_t*) xp;
    y = (fpair_t*) yp;
    if(x->etime.tv_sec == y->etime.tv_sec) {
        if(x->etime.tv_usec == y->etime.tv_usec) {
            return x->chan - y->chan;
        }
        else {
            return x->etime.tv_usec - y->etime.tv_usec;
        }
    }
    else {
        return x->etime.tv_sec - y->etime.tv_sec;
    }
}

show_data()
{
    int i;

    for(i=0;i<img_pool_count;i++) {
        printf("image %6d: %p\n", i, &img_pool[i]);
    }
    for(i=0;i<img_pool_count;i++) {
        printf("fpair %6d: %d.%06d %2d %p\n", i, 
        fpairs[i].etime.tv_sec, fpairs[i].etime.tv_usec,
        fpairs[i].chan, fpairs[i].imgpos);
    }
}

sentry_imgrec* get_free_slot()
{
    static int i=0;
    sentry_imgrec *ret;
    int c;
    
    ret = NULL;
    c = 0;
    while(c<IMG_POOL_MAX) {
        if(img_pool[i].recv_sec<=0) {
            ret = &img_pool[i];
            goto out;
        }
        i = (i+1)%IMG_POOL_MAX;
        c++;
    }

out:
#if 0
    if(ret==NULL) 
        i = -1;
    printf("new slot %7d %p\n", i, ret);
#endif
    return ret;
}

int
parse_sentry_packedfile(char *fname, int *rc, int chan, time_t limit)
{
    unsigned char tokenbuf[BUFSIZ];
    int chk;
    int fd;
    off_t pos;
    token_t rtoken, ctoken;
    int skiplen;
    int nr;
    int tokenlen;
    int32_t *p;
    unsigned char headbuf[BUFSIZ];
    sentry_imgrec rtmpimg;
    sentry_imgrec *ctmpimg;
    sentry_imgrec _ctmpimg;
    sentry_imgrec *cimg;
    unsigned char *body_ptr;
    int rc_start;

    tokenlen = 6 + sizeof(int32_t)*2;
    rc_start = *rc;

    if(debug>=10000) {
        fprintf(stderr, "reading target path '%s'\n", fname);
    }

    fd = open(fname, O_RDONLY);
    if(fd<0) {
        if(errno==ENOENT) {
            goto out;
        }
        fprintf(stderr, "fail open '%s' (%d)\n", fname, errno);
        goto fail;
    }

    pos = lseek(fd, 0, SEEK_SET);
    if(pos<0) {
        fprintf(stderr, "fail lseek '%s' (%d)\n", fname, errno);
        goto fail;
    }

/*
    rc = 0;
*/
    while(1) {
        memset(tokenbuf, 0xff, tokenlen);

try_read1:
        nr = read(fd, tokenbuf, tokenlen);
        if(nr<0) {
            fprintf(stderr, "fail read '%s' (%d)\n", fname, errno);
            goto suspend;
        }
        if(nr==0) {
            if(debug>=1000)
            fprintf(stderr, "       done\n");
            break;
        }
        if(nr<tokenlen) {
            fprintf(stderr, "      lack of size, read %d expect %d. give up\n",
                nr, tokenlen);
            break;
        }

        if(debug>=20000) {
            int i;
            int v;

            fprintf(stderr, "   ");
            for(i=0;i<nr;i++) {
                fprintf(stderr, "%02X ", i);
            }
            fprintf(stderr, " \n");
            
            fprintf(stderr, "  <");
            for(i=0;i<nr;i++) {
                fprintf(stderr, "%02x ", tokenbuf[i]);
            }
            fprintf(stderr, ">\n");
        }

        memset(&rtoken, 0, sizeof(rtoken));

        strncpy(rtoken.magick,  &tokenbuf[0],  6);
        memcpy(&rtoken.ver,   &tokenbuf[6],  4);
        memcpy(&(rtoken.len),   &tokenbuf[10], sizeof(int32_t));

#if 0
        fprintf(stderr, " %4d: <%06d> r \"%c%c%c%c%c%c\" %2d %4d\n",
            *rc, (long)pos,

            rtoken.magick[0],
            rtoken.magick[1],
            rtoken.magick[2],
            rtoken.magick[3],
            rtoken.magick[4],
            rtoken.magick[5],
            rtoken.ver, rtoken.len);
#endif



        chk = strncmp(rtoken.magick, "SENTRY", 6);
        if(chk!=0) {
            fprintf(stderr, "ignore magic '%c%c%c%c%c' %d\n",
                rtoken.magick[0],
                rtoken.magick[1],
                rtoken.magick[2],
                rtoken.magick[3],
                rtoken.magick[4],
                rtoken.magick[6],
                chk);
            goto suspend;
        }

        strncpy(ctoken.magick, rtoken.magick, 6);
        ctoken.ver = ntohl(rtoken.ver);
        ctoken.len = ntohl(rtoken.len);
    
        skiplen = ctoken.len;

#if 0
        fprintf(stderr, " %4d: <%06d> c \"%c%c%c%c%c%c\" %2d %4d\n",
            *rc, (long)pos,
            ctoken.magick[0],
            ctoken.magick[1],
            ctoken.magick[2],
            ctoken.magick[3],
            ctoken.magick[4],
            ctoken.magick[5],
            ctoken.ver, ctoken.len);
#endif

try_read2:
#if 0
        nr = read(fd, headbuf, sizeof(sentry_imgrec));
#endif
        nr = read(fd, headbuf, SENTRY_SHM_HEADER_SIZE);
        if(nr<0) {
            if(errno==EINTR) {
                goto try_read2;
            }
            fprintf(stderr, "cannot read header (%d)\n", errno);
            goto suspend;
        }
        if(nr==0) {
            break;
        }

        skiplen -= nr;
        
        memcpy(&rtmpimg, headbuf, sizeof(sentry_imgrec));
#if 0
        /* unreadble data (because it is not host order) */
        if(debug>=1000)
            show_imgattr(stderr, &rtmpimg);
#endif

        ctmpimg = &_ctmpimg;

        ntoh_img(ctmpimg, &rtmpimg);
        if(debug>=1000)
            show_imgattr(stderr, ctmpimg);

        if(ctmpimg->length>MAX_IMGSIZE || ctmpimg->recv_sec<limit) {
            body_ptr = NULL;

#if 0
            ctmpimg->body_len = 0;
            ctmpimg->body = NULL;
#endif
#if 0
            ctmpimg = NULL;
#endif
        }
        else {
            body_ptr = (unsigned char*)malloc(ctmpimg->length);
            img_heap_use += ctmpimg->length;
try_bodyread1:
            nr = read(fd, body_ptr, ctmpimg->length);
            if(nr<0) {
                if(errno==EINTR) {
                    goto try_bodyread1;
                }
                fprintf(stderr, "cannot read header (%d)\n", errno);
                goto suspend;
            }

#if 0
            fprintf(stderr, "read body %d\n", ctmpimg->length);
#endif
            skiplen -= ctmpimg->length;

            ctmpimg->body_len = ctmpimg->length;
            ctmpimg->body = body_ptr;

            cimg = get_free_slot();
            if(cimg==NULL) {
                fprintf(stderr,
                    "no more free slot %d (use %ditmes, %dbytes)\n",
                    IMG_POOL_MAX, img_pool_count, img_heap_use);
                fflush(stderr);
            }
            else {
                *cimg = *ctmpimg;
            }

            fpairs[img_pool_count].etime.tv_sec     = ctmpimg->recv_sec;
            fpairs[img_pool_count].etime.tv_usec    = ctmpimg->recv_usec;
            fpairs[img_pool_count].chan             = chan;
            fpairs[img_pool_count].imgpos           = cimg;

 
            img_pool_count++;
        }

#if 0
        fprintf(stderr, "  r length %d, size %dx%d\n",
            ntohl(rtmpimg.length),
            ntohl(rtmpimg.width), ntohl(rtmpimg.height));
#endif


#if 0
        fprintf(stderr, " %4d: <%06d> c \"%c%c%c%c%c%c\" %2d %4d\n",
            *rc, (long)pos,
            ctoken.magick[0],
            ctoken.magick[1],
            ctoken.magick[2],
            ctoken.magick[3],
            ctoken.magick[4],
            ctoken.magick[5],
            ctoken.ver, ctoken.len);
#endif
#if 0
        fprintf(stderr, "  c length %d, size %dx%d\n",
            (ctmpimg.length),
            (ctmpimg.width), (ctmpimg.height));
#endif


        if(debug>=1000 && ctmpimg) 
        fprintf(stderr, " %4d: %d.%06d %d.%06d %s %s %03dx%03d\n",
            *rc,
            ctmpimg->send_sec, ctmpimg->send_usec,
            ctmpimg->recv_sec, ctmpimg->recv_usec,
            assoc_v(format_alist, ctmpimg->format),
            assoc_v(color_alist, ctmpimg->corder),
            ctmpimg->width, ctmpimg->height);
    

        if(debug>=1000)
            show_imgattr(stderr, ctmpimg);


        (*rc)++;

        if(skiplen) {
            pos = lseek(fd, skiplen, SEEK_CUR);
            if(pos<0) {
                fprintf(stderr, "fail lseek '%s' (%d)\n", fname, errno);
                break;
            }
        }
    }

suspend:
    close(fd);

    if(debug>=500)
        fprintf(stderr, "%4d images [%4d,%4d) are stored in '%s'\n",
            *rc-rc_start, rc_start, *rc, fname);

    
    goto out;
    
fail:

out:
    if(debug>=500)
        fprintf(stderr, "img_pool_count %d/%d, img_heap_use %d\n",
            img_pool_count, IMG_POOL_MAX, img_heap_use);

    return 0;
}

scan_frames(char *prefix, int chan, time_t ref, time_t range, time_t limit)
{
    int i;
    int nc;
    int start, end;
    int t;
    int diff;
    char tmppath[BUFSIZ];
    int kk;

    if(debug>=10) {
        printf("scan_frames: chan %d, range %d -%d, path '%s'\n",
            chan, ref, range, prefix);
    }

    start = ref - range;
    end   = ref;

    if(debug>=10) {
        printf(" [%d..%d]\n", start, end);
    }
    
    kk = 0;
    i = 0;
    for(t=start; t<=end; t+=60) {
        diff = ref - t ;
#if 0
        printf("%6d: %d %5dm %2dd_%04dm_%02ds\n",
            i, t, diff/60, diff/86400, (diff%86400)/60, diff%60);
#endif
        conv_filepath(tmppath, prefix, 1, t);
        parse_sentry_packedfile(tmppath, &kk, chan, limit);

#if 0
        printf("%6d: %d %5dm %2dd_%04dm_%02ds '%s'\n",
            i, t, diff/60, diff/86400, (diff%86400)/60, diff%60, tmppath);
#endif

        i++;
    }

#if 0
    show_data();
#endif
}

#ifdef TEST_SFSCAN

main(int argc, char **argv)
{
    int now;

    debug=700;
    
    now = time(NULL);
    scan_frames("/var/tmp/sen_bank/c1", 1, now);
    scan_frames("/var/tmp/sen_bank/c2", 2, now);

    show_data();

    qsort(fpairs, img_pool_count, sizeof(fpair_t), fpair_cmp);

    show_data();
}

#endif /* TEST_SFSCAN */
